// Cache-Agnostic Spinlock
//
// * 4 processes
// * crit dist  : p(t=1) = 1/1
//      scaled  : p(t=0) = 1/1
// * ncrit dist : p(t=8) = 1/2, p(t=10) = 1/2
//       scaled : p(t=18) = 1/2, p(t=14) = 1/2
// * init dist  : ncrit dist
dtmc

const n     = 4;
const t_max = 18;



formula race         = (loc_obs = wait | loc_sym_2 = wait | loc_sym_3 = wait | loc_sym_4 = wait)
                       & (lock = free | (loc_obs = crit & t_obs = 0) | (loc_sym_2 = crit & t_sym_2 = 0) | (loc_sym_3 = crit & t_sym_3 = 0) | (loc_sym_4 = crit & t_sym_4 = 0));
formula obs_requests = loc_obs = wait & t_obs = 0;
formula obs_failed   = lock != assign_obs & loc_obs = wait & t_obs = 1;
formula obs_spins    = lock != assign_obs & loc_obs = wait & t_obs > 0;
formula obs_holds    = lock = assign_obs;
formula obs_releases = loc_obs = crit & t_obs = 0;

label "race"         = race;
label "obs_requests" = obs_requests;
label "obs_failed"   = obs_failed;
label "obs_spins"    = obs_spins;
label "obs_holds"    = obs_holds;
label "obs_releases" = obs_releases;

rewards "obs_spins"
    obs_spins : 1;
endrewards

rewards "waiting_processes"
    true : (loc_obs = wait ? 1 : 0)
           + (loc_sym_2 = wait ? 1 : 0)
           + (loc_sym_3 = wait ? 1 : 0)
           + (loc_sym_4 = wait ? 1 : 0);
endrewards



const ncrit  = 2;
const wait   = 0;
const crit   = 1;

module Process_obs
    t_obs   : [0..t_max] init 0;
    loc_obs : [0..2]     init crit;

    // tick in ncrit
    [tick] loc_obs = ncrit & t_obs > 0
        -> (t_obs' = t_obs - 1);
    [tick] loc_obs = ncrit & t_obs = 0
        -> (loc_obs' = wait);
    // wait for lock at least one tick
    [tick] loc_obs = wait & (t_obs = 0 | lock != assign_obs)
        -> (t_obs' = 1);
    // grab lock
    [tick] loc_obs = wait & t_obs = 1 & lock = assign_obs
        -> 1/1 : (loc_obs' = crit) & (t_obs' = 0);
    // tick in crit
    [tick] loc_obs = crit & t_obs > 0
        -> (t_obs' = t_obs - 1);
    // release lock
    [tick] loc_obs = crit & t_obs = 0
        -> 1/2 : (loc_obs' = ncrit) & (t_obs' = 18)
         + 1/2 : (loc_obs' = ncrit) & (t_obs' = 14);
endmodule



module Process_sym_2
    t_sym_2   : [0..t_max] init 0;
    loc_sym_2 : [0..2]     init crit;

    // tick in ncrit
    [tick] loc_sym_2 = ncrit & t_sym_2 > 0
        ->  (t_sym_2' = t_sym_2 - 1);
    [tick] loc_sym_2 = ncrit & t_sym_2 = 0
        -> (loc_sym_2' = wait);
    // wait for lock at least one tick
    [tick] loc_sym_2 = wait & lock != assign_2
        ->  true;
    // grab lock
    [tick] loc_sym_2 = wait & lock = assign_2
        -> 1/1 : (loc_sym_2' = crit) & (t_sym_2' = 0);
    // tick in crit
    [tick] loc_sym_2 = crit & t_sym_2 > 0
        ->  (t_sym_2' = t_sym_2 - 1);
    [tick] loc_sym_2 = crit & t_sym_2 = 0
        -> 1/2 : (loc_sym_2' = ncrit) & (t_sym_2' = 18)
         + 1/2 : (loc_sym_2' = ncrit) & (t_sym_2' = 14);
endmodule



module Process_sym_3 = Process_sym_2
                        [
                            t_sym_2   = t_sym_3,
                            loc_sym_2 = loc_sym_3,
                            assign_2  = assign_3
                        ]
endmodule



module Process_sym_4 = Process_sym_2
                        [
                            t_sym_2   = t_sym_4,
                            loc_sym_2 = loc_sym_4,
                            assign_2  = assign_4
                        ]
endmodule



const free       = 0;
const assign_obs = 1;
const assign_2   = 2;
const assign_3   = 3;
const assign_4   = 4;

module Lock
    lock : [0..4] init free;

    // self loop in unlock
    [tick] lock = free & !(loc_obs = wait | loc_sym_2 = wait | loc_sym_3 = wait | loc_sym_4 = wait)
        -> true;
    // from unlock to assign_obs
    [tick] lock = free & loc_obs = wait
        -> (lock' = assign_obs);
    // self loop in assign_obs
    [tick] lock = assign_obs & !(loc_obs = crit & t_obs = 0)
        -> true;
    //from assign_obs to other locks
    [tick] lock = assign_obs & loc_obs = crit & t_obs = 0 & loc_sym_2 = wait
        -> (lock' = assign_2);
    [tick] lock = assign_obs & loc_obs = crit & t_obs = 0 & loc_sym_3 = wait
        -> (lock' = assign_3);
    [tick] lock = assign_obs & loc_obs = crit & t_obs = 0 & loc_sym_4 = wait
        -> (lock' = assign_4);
    // from assign_obs to free
    [tick] lock = assign_obs & loc_obs = crit & t_obs = 0 & !(loc_sym_2 = wait | loc_sym_3 = wait | loc_sym_4 = wait)
        -> (lock' = free);
    // from unlock to assign_2
    [tick] lock = free & loc_sym_2 = wait
        -> (lock' = assign_2);
    // self loop in assign_2
    [tick] lock = assign_2 & !(loc_sym_2 = crit & t_sym_2 = 0)
        -> true;
    // from assign_2 to other locks
    [tick] lock = assign_2 & loc_sym_2 = crit & t_sym_2 = 0 & loc_obs = wait
        -> (lock' = assign_obs);
    [tick] lock = assign_2 & loc_sym_2 = crit & t_sym_2 = 0 & loc_sym_3 = wait
        -> (lock' = assign_3);
    [tick] lock = assign_2 & loc_sym_2 = crit & t_sym_2 = 0 & loc_sym_4 = wait
        -> (lock' = assign_4);
    // from assign_2 to free
    [tick] lock = assign_2 & loc_sym_2 = crit & t_sym_2 = 0 & !(loc_obs = wait | loc_sym_3 = wait | loc_sym_4 = wait)
        -> (lock' = free);
    // from unlock to assign_3
    [tick] lock = free & loc_sym_3 = wait
        -> (lock' = assign_3);
    // self loop in assign_3
    [tick] lock = assign_3 & !(loc_sym_3 = crit & t_sym_3 = 0)
        -> true;
    // from assign_3 to other locks
    [tick] lock = assign_3 & loc_sym_3 = crit & t_sym_3 = 0 & loc_obs = wait
        -> (lock' = assign_obs);
    [tick] lock = assign_3 & loc_sym_3 = crit & t_sym_3 = 0 & loc_sym_2 = wait
        -> (lock' = assign_2);
    [tick] lock = assign_3 & loc_sym_3 = crit & t_sym_3 = 0 & loc_sym_4 = wait
        -> (lock' = assign_4);
    // from assign_3 to free
    [tick] lock = assign_3 & loc_sym_3 = crit & t_sym_3 = 0 & !(loc_obs = wait | loc_sym_2 = wait | loc_sym_4 = wait)
        -> (lock' = free);
    // from unlock to assign_4
    [tick] lock = free & loc_sym_4 = wait
        -> (lock' = assign_4);
    // self loop in assign_4
    [tick] lock = assign_4 & !(loc_sym_4 = crit & t_sym_4 = 0)
        -> true;
    // from assign_4 to other locks
    [tick] lock = assign_4 & loc_sym_4 = crit & t_sym_4 = 0 & loc_obs = wait
        -> (lock' = assign_obs);
    [tick] lock = assign_4 & loc_sym_4 = crit & t_sym_4 = 0 & loc_sym_2 = wait
        -> (lock' = assign_2);
    [tick] lock = assign_4 & loc_sym_4 = crit & t_sym_4 = 0 & loc_sym_3 = wait
        -> (lock' = assign_3);
    // from assign_4 to free
    [tick] lock = assign_4 & loc_sym_4 = crit & t_sym_4 = 0 & !(loc_obs = wait | loc_sym_2 = wait | loc_sym_3 = wait)
        -> (lock' = free);
endmodule
